from sklearn.decomposition import PCA
import matplotlib
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.axes3d import Axes3D
from sklearn.cluster import KMeans
import matplotlib.cm as cm
import numpy as np
import pandas as pd
from typing import List
from matplotlib.axes import Axes

class DataAnalyzer:
    def __init__(self, n_components: int, metrics: List[str]):
        try:
            if n_components < 2 or n_components > 3:
                raise ValueError('n_components should be 2 or 3.')

            self.n_components = n_components
            self.metrics = metrics
            self.axis = ['x', 'y', 'z']
            self.cmap = self._getCM()

        except ValueError as e:
            print(e)
    
    def applyPCA(self, dataframe: pd.DataFrame):
        pca = PCA(n_components=self.n_components)
        fitted = pca.fit_transform(dataframe[self.metrics])
        dataframe[self.axis[:self.n_components]] = pd.DataFrame(fitted, index=dataframe.index)
    
    def applyKMeans(self, dataframe: pd.DataFrame, n_clusters: int) -> KMeans:
        kmeans = KMeans(n_clusters=n_clusters)
        kmeans.fit(dataframe[self.axis[:self.n_components]])

        return kmeans

    def _getCM(self):
        cmap = cm.jet
        cmaplist = [cmap(i) for i in range(cmap.N)]
        cmap = cmap.from_list('cmap', cmaplist, cmap.N)

        return cmap

    def plot(self, dataframe: pd.DataFrame, labels: List[int], centroids: List[List[float]]):
        if self.n_components == 2:
            ax = plt.subplot(1, 1, 1)
            # scat = ax.scatter(dataframe['x'], dataframe['y'], c=labels, cmap=self.cmap, label=labels)
        else:
            fig = plt.figure()
            ax = Axes3D(fig)
            # scat = ax.scatter(dataframe['x'], dataframe['y'], dataframe['z'], c=labels, cmap=self.cmap, label=labels)

        #legend = ax.legend(*scat.legend_elements(), loc='upper right', title='Clusters')
        #ax.add_artist(legend)
        # , marker='°'
        arr = np.arange(len(centroids))
        ax.scatter(centroids[:, 0], centroids[:, 1], c=arr, cmap=self.cmap, marker='*', s=400, lw=1.5, edgecolor=(0, 0, 0, 1))

        # ax.set_xlim([-1000000, 8000000])
        # ax.set_ylim([0, 2600])
        # if (self.n_components == 3):
        #     ax.set_zlim([-80, 80])

        ax.set_xticklabels([])
        ax.set_yticklabels([])
        if (self.n_components == 3):
            ax.set_zticklabels([])
        return ax, plt

    def addPatientPoint(self, dataframe: pd.DataFrame, ax: Axes, processInstanceId: str, lastOrKM: bool, kmeans: KMeans):
        #If lastOrKM is true it plots the cluster of all patient's data
        patientData = dataframe[dataframe['processInstanceId'] == processInstanceId]

        if lastOrKM:
            kmeansPatient = KMeans(n_clusters=1)
            kmeansPatient.fit(patientData[self.axis[:self.n_components]])
            pointToPlot = kmeansPatient.cluster_centers_[0]
        else:
            pointToPlot = patientData.tail(1)[self.axis[:self.n_components]].to_numpy()[0]

        if self.n_components == 2:
            ax.scatter(pointToPlot[0], pointToPlot[1], c=kmeans.predict(pointToPlot.reshape(1, -1)), s=250, marker='x')
        else:
            ax.scatter(pointToPlot[0], pointToPlot[1], pointToPlot[2], c=kmeans.predict(pointToPlot.reshape(1, -1)), s=250, marker='x', cmap=self.cmap)
        
        return pointToPlot

    def addAllPatientPoints(self, dataframe: pd.DataFrame, ax: Axes, processInstanceId: str):
        patientData = dataframe[dataframe['processInstanceId'] == processInstanceId]

        data = patientData[self.axis[:self.n_components]].to_numpy()
        
        if self.n_components == 2:
            x = data[:, 0]
            y = data[:, 1]

            ax.quiver(x[:-1:4], y[:-1:4], x[1::4] - x[:-1:4], y[1::4] - y[:-1:4])
            #ax.plot(data[:, 0], data[:, 1], c='green')
        else:
            x = data[:, 0]
            y = data[:, 1]
            z = data[:, 2]

            ax.quiver(x[:-1], y[:-1], z[:-1], x[1:] - x[:-1], y[1:] - y[:-1], z[1:] - z[:-1], length=10, normalize=True, arrow_length_ratio=0.9)
